#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include <process.h>
#include <windows.h>

#include "ini.h"
#include "lib_egpc_b4s1.h"

#define LIB_FCANBUS "lib_egpc_b4s1.dll"
#define EXIT_EVENT  0xFF

HMODULE hDLL = NULL;
HANDLE  TRD_send = NULL;
static DWORD64 rxCnt = 0;
static DWORD64 errorCnt1 = 0;
static DWORD64 errorCnt2 = 0;

EGPC_OPEN           EGPCOpen;
EGPC_CLOSE          EGPCClose;
EGPC_SET_BAUDRATE   EGPCSetBaudRate;
EGPC_GET_BAUDRATE   EGPCGetBaudRate;
EGPC_SET_FILTER     EGPCSetFilter;
EGPC_CLEAR_FILTER   EGPCClearFilter;
EGPC_RECEIVE        EGPCReceive;
EGPC_SEND           EGPCSend;
EGPC_RESET          EGPCReset;
EGPC_SHOW_VER       EGPCShowVer;

void CALLBACK cb_receive_can_info(long error, CanFrameInfor* message);
static void get_ini_setting (void);
static int  handler_ini (void* user, const char* section, const char* name, const char* value);
static void print_config (void);
static void load_dll (void);
static void demo_egpc (char *comport, DWORD setBaudRate);
static void start_egpc(void);
static BOOL end_egpc (DWORD type);
static DWORD WINAPI TRD_send_fx (LPVOID ptr);
static void send_group (void);


/* setup.ini */
/*----------------------------------------------------------*/
typedef struct
{
    /* [setting] */
    char  comport[7];
    DWORD baudrate;
    int   send_cnt_once;
    int   sleep_interval;
    bool  show_send_rtn;
    bool  re_send;
    bool  print_rx_raw_data;

    /* [data] */
    int data_length;
    unsigned char test_data[CANBUS_MAX_DATA_SIZE];
    bool rtr;

    /* [id] */
    bool id_increase;
    bool id_type;
    unsigned int id;

    /* [filter] */
    bool filter_1;
    bool filter_2;
    int filter_1_pattern;
    int filter_1_mask;
    int filter_2_pattern;
    int filter_2_mask;

} INI_ITEM;

INI_ITEM  ini_item;

/*---------------------------------------------------------------------------------*/
int main (void)
{
    char ver_info[VER_LEN];
    
    load_dll();

    EGPCShowVer(ver_info);

    printf("EGPC library version: %s\n", ver_info);

    get_ini_setting();

    print_config();

    printf("\n**************** set CAN initialization ****************\n");
    demo_egpc(ini_item.comport, ini_item.baudrate);
    printf("\n********************************************************\n\n");
    start_egpc();

    return 0;
} /* END: main() */


/*---------------------------------------------------------------------------------*/
static void demo_egpc (char *comport, DWORD setBaudRate)
{
    long rtn;

    /* ----- EGPOpen() ----- */
	printf("EGPCOpen()\n");
    rtn = EGPCOpen(comport);
    if(rtn)
    {
        printf("Open %s failed (rtn=%lx) !\n", comport, rtn);
        exit(1);
    }
    else
    {
        printf("Open %s successfully !\n", comport);
        printf("==============================================\n");
    }

    /* ----- EGPCSetBaudRate() ----- */
	printf("EGPCSetBaudRate()\n");
    rtn = EGPCSetBaudRate(comport, setBaudRate);
    if(rtn)
    {
        printf("EGPC Set BaudRate failed (rtn=%lx) !\n", rtn);
        exit(1);
    }
    else
    {
        printf("EGPC Set BaudRate successfully !\n");
        printf("==============================================\n");
    }

    /* ----- EGPCGetBaudRate() ----- */
	printf("EGPCGetBaudRate()\n");
    DWORD getBaudRate = 0;
    rtn = EGPCGetBaudRate(comport, &getBaudRate);
    if(rtn)
    {
        printf("EGPC Get BaudRate failed (rtn=%lx) !\n", rtn);
        exit(1);
    }
    else
    {
        printf("EGPC BaudRate: %lu !\n", getBaudRate);
        printf("==============================================\n");
    }



    /* ----- EGPCSetFilter() ----- */
    if(ini_item.filter_1 || ini_item.filter_2)
	    printf("EGPCSetFilter()\n");
    
    if(ini_item.filter_1)
    {

        rtn = EGPCSetFilter(comport, ini_item.filter_1_pattern, ini_item.filter_1_mask);
        if(rtn < 0)
        {
            printf("EGPC Set Filter failed (rtn=%lx) !\n", rtn);
            exit(1);
        }
        else
        {
            printf("EGPC Set Filter successfully (filter count: %ld) !\n", rtn+1);
            printf("==============================================\n");
            printf("Set filter_1 is: 0x%x\n\n",ini_item.filter_1_pattern);
        }
    }

    if(ini_item.filter_2)
    {

        rtn = EGPCSetFilter(comport, ini_item.filter_2_pattern, ini_item.filter_2_mask);
        if(rtn < 0)
        {
            printf("EGPC Set Filter failed (rtn=%lx) !\n", rtn);
            exit(1);
        }
        else
        {
            printf("EGPC Set Filter successfully (filter count: %ld) !\n", rtn+1);
            printf("==============================================\n");
            printf("Set filter_2 is: 0x%x\n",ini_item.filter_2_pattern);
        }
    }
        
    // TO-DO: rtn, filter total count - 1 (0-14, MAX: 15)
    

#if 0

    /* ----- EGPCClearFilter() ----- */
	printf("EGPCClearFilter()\n");
    rtn = EGPCClearFilter(comport);
    printf("rtn is rtn=%lx !\n", rtn);
    if(rtn)
    {
        printf("EGPC Clear Filter failed (rtn=%lx) !\n", rtn);
        exit(1);
    }
    else
    {
        printf("EGPC Clear Filter successfully !\n");
        printf("==============================================\n");
    }
    
#endif

    /* ----- EGPCReset() ----- */
	printf("EGPCReset()\n");
    rtn = EGPCReset(comport);
    if(rtn)
    {
        printf("EGPC Reset failed (rtn=%lx) !\n", rtn);
        exit(1);
    }
    else
    {
        printf("EGPC Reset successfully !\n");
        printf("==============================================\n");
    }

} /* END: demo_egpc() */

/*---------------------------------------------------------------------------------*/
static void start_egpc(void)
{
    long rtn;

    /* ----- EGPCReceive() ----- */
    printf("EGPCReceive()\n");

    rtn = EGPCReceive(ini_item.comport, cb_receive_can_info);
    if(rtn)
    {
        printf("EGPC Receive failed (rtn=%lx) !\n", rtn);
        exit(1);
    }
    else
    {
        printf("EGPC Receive successfully !\n");
        printf("==============================================\n");
    }

    TRD_send = CreateThread(NULL, 0, TRD_send_fx, NULL, 0, NULL);
    SetConsoleCtrlHandler((PHANDLER_ROUTINE) end_egpc, TRUE);
    WaitForSingleObject(TRD_send, INFINITE);

} /* END: start_egpc() */


/*---------------------------------------------------------------------------------*/
static BOOL end_egpc (DWORD type)
{
    if(type == CTRL_C_EVENT || type == EXIT_EVENT)
    {

        printf("\nExit the program ...\n");

        TerminateThread(TRD_send, 0);

        /* ----- EGPCClose() ----- */
        EGPCClose(ini_item.comport);

        exit(0);
    }

    /* prevent windows compiler warning */
    return TRUE;

} /* END: end_egpc() */


/*---------------------------------------------------------------------------------*/
static DWORD WINAPI TRD_send_fx (LPVOID ptr)
{
    while (1)
    {
        printf("\n ------------------------- Press 'Enter' to send data, 'Ctrl + C' to exit -------------------------\n");
        getchar();

        send_group();
    }

    return 0;

} /* END: TRD_send_fx() */


/*---------------------------------------------------------------------------------*/
static void send_group (void)
{
    static int      cnt_record = 0;
    int             i;
    long            rtn;
    CanFrameInfor   frame_send;

    memset(&frame_send, 0, sizeof(CanFrameInfor));

    if(ini_item.id_type)
        frame_send.type = CP_29Bit;
    else
        frame_send.type = CP_11Bit;

    frame_send.rtr = ini_item.rtr;
    frame_send.data_len = ini_item.data_length;
    frame_send.id = ini_item.id;
    memcpy(frame_send.data, ini_item.test_data, CANBUS_MAX_DATA_SIZE);

    for(i=1; i<=ini_item.send_cnt_once; i++)
    {
        if(ini_item.id_increase == true)
        {
            frame_send.id = (unsigned int) i;
        }

        do
        {
            /* ----- EGPCSend() ----- */
            rtn = EGPCSend(ini_item.comport, &frame_send);

            if (true == ini_item.show_send_rtn)
            {
                printf("%4d. send rtn=%lx\n", i, rtn);
            }

            if (rtn == 0)
            {
                break;
            }

        } while (ini_item.re_send);

        if(rtn)
        {
#if 1
            printf("Please check connect or Setup.ini information (Ex: BaudRate etc.) \n");
            system("pause");
            end_egpc(EXIT_EVENT);
#else
            /* ----- EGPCReset() ----- */
            /* If an error occurs, you can reset by EGPCReset() */
            long rtn = EGPCReset(ini_item.comport);
            printf("EGPCReset(rtn=%lx) !\n", rtn);
#endif
        }

        if(ini_item.sleep_interval)
            Sleep(ini_item.sleep_interval);
    }

    cnt_record += ini_item.send_cnt_once;
    printf("\nSend END -----------------------> cnt: %d, cumulation: %d\n\n", ini_item.send_cnt_once, cnt_record);


} /* END: send_group() */


/*---------------------------------------------------------------------------------*/
static void get_ini_setting (void)
{
    if(ini_parse("setup.ini", handler_ini, &ini_item) < 0)
    {
        printf("Open setup.ini failed\n");
        exit(1);
    }

} /* END: get_ini_setting() */


/*---------------------------------------------------------------------------------*/
static int handler_ini (void *user, const char *section, const char *name, const char *value)
{
    #define MATCH(s, n) strcmp(section, s) == 0 && strcmp(name, n) == 0

    INI_ITEM *p_ini = (INI_ITEM *)user;

    /* [setting] section */
    if(MATCH("setting", "comport"))
    {
        strcpy(p_ini->comport, value);
        p_ini->comport[strlen(value)] = '\0';
    }
    else if(MATCH("setting", "baudrate"))
    {
        p_ini->baudrate = (DWORD) atoi(value);
    }
    else if(MATCH("setting", "send_cnt_once"))
    {
        p_ini->send_cnt_once = (int) atoi(value);
    }
    else if(MATCH("setting", "sleep_interval"))
    {
        p_ini->sleep_interval = (int) atoi(value);
    }
    else if(MATCH("setting", "show_send_rtn"))
    {
        p_ini->show_send_rtn = (bool) atoi(value);
    }
    else if(MATCH("setting", "re_send"))
    {
        p_ini->re_send = (bool) atoi(value);
    }
    else if(MATCH("setting", "print_rx_raw_data"))
    {
        p_ini->print_rx_raw_data = (bool) atoi(value);
    }

    /* [data] section */
    else if(MATCH("data", "data_length"))
    {
        p_ini->data_length = (int) atoi(value);
    }
    else if(MATCH("data", "test_data_0"))
    {
        sscanf(value, "0x%hhx", &(p_ini->test_data)[0]);
    }
    else if(MATCH("data", "test_data_1"))
    {
        sscanf(value, "0x%hhx", &(p_ini->test_data)[1]);
    }
    else if(MATCH("data", "test_data_2"))
    {
        sscanf(value, "0x%hhx", &(p_ini->test_data)[2]);
    }
    else if(MATCH("data", "test_data_3"))
    {
        sscanf(value, "0x%hhx", &(p_ini->test_data)[3]);
    }
    else if(MATCH("data", "test_data_4"))
    {
        sscanf(value, "0x%hhx", &(p_ini->test_data)[4]);
    }
    else if(MATCH("data", "test_data_5"))
    {
        sscanf(value, "0x%hhx", &(p_ini->test_data)[5]);
    }
    else if(MATCH("data", "test_data_6"))
    {
        sscanf(value, "0x%hhx", &(p_ini->test_data)[6]);
    }
    else if(MATCH("data", "test_data_7"))
    {
        sscanf(value, "0x%hhx", &(p_ini->test_data)[7]);
    }
    else if(MATCH("data", "rtr"))
    {
        p_ini->rtr = (bool) atoi(value);
    }

    /* [id] section */
    else if(MATCH("id", "id_increase"))
    {
        p_ini->id_increase = (bool) atoi(value);
    }
    else if(MATCH("id", "id_type"))
    {
        p_ini->id_type = (bool) atoi(value);
    }
    else if(MATCH("id", "id"))
    {
        sscanf(value, "%x" , &p_ini->id);
    }

    /*[filter] section*/
    else if(MATCH("filter", "filter_1_usage"))
    {
        p_ini->filter_1 = (bool) atoi(value);
    }
    else if(MATCH("filter", "filter_2_usage"))
    {
        p_ini->filter_2 = (bool) atoi(value);
    }
    else if(MATCH("filter", "filter_1_pattern"))
    {
        sscanf(value, "%x" , &p_ini->filter_1_pattern);
    }
    else if(MATCH("filter", "filter_1_mask"))
    {
        sscanf(value, "%x" , &p_ini->filter_1_mask);
    }
    else if(MATCH("filter", "filter_2_pattern"))
    {
        sscanf(value, "%x" , &p_ini->filter_2_pattern);
    }
    else if(MATCH("filter", "filter_2_mask"))
    {
        sscanf(value, "%x" , &p_ini->filter_2_mask);
    }

    /* unknown section/name, error */
    else
    {
        return 0;
    }

    return 1;

} /* END: handler_ini() */


/*---------------------------------------------------------------------------------*/
static void print_config (void)
{
    printf("\nConfig\n");
    printf("----------------------------------\n");
    printf("comport: %s\n", ini_item.comport);
    printf("baudrate: %lu\n", ini_item.baudrate);
    printf("send_cnt_once: %d\n", ini_item.send_cnt_once);
    printf("sleep_interval: %d\n", ini_item.sleep_interval);
    printf("show_send_rtn: %s\n", ini_item.show_send_rtn ? "on" : "off");
    printf("re_send: %s\n", ini_item.re_send ? "on" : "off");
    printf("print_rx_raw_data: %s\n", ini_item.print_rx_raw_data ? "on" : "off");
    printf("data_length: %d\n", ini_item.data_length);
    printf("test_data: ");
    for (int i = 0; i<CANBUS_MAX_DATA_SIZE; i++)
    {
        printf("%02X ", ini_item.test_data[i]);
    }
    printf("\n");
    printf("rtr: %s\n", ini_item.rtr ? "enable" : "disable");
    printf("id_increase: %s\n", ini_item.id_increase ? "on" : "off");
    printf("id_type: %s\n", ini_item.id_type ? "29Bit" : "11Bit");
    printf("id: 0x%x\n", ini_item.id);
    printf("\n");
    printf("filter_1: %s\n", ini_item.filter_1 ? "on" : "off");
    printf("filter_2: %s\n", ini_item.filter_2 ? "on" : "off");
    if(ini_item.filter_1) 
        printf("Set filter_1 is: 0x%x\n", ini_item.filter_1_pattern);
    if(ini_item.filter_2)
        printf("Set filter_2 is: 0x%x\n", ini_item.filter_2_pattern);
    printf("----------------------------------\n\n");


} /* END: print_config() */


/*---------------------------------------------------------------------------------*/
static void load_dll (void)
{
    hDLL = LoadLibrary(LIB_FCANBUS);

    if(hDLL == NULL)
    {
        printf("Load %s failed!(%ld)\n", LIB_FCANBUS, GetLastError());
        exit(1);
    }
    else
    {
        EGPCOpen         = (EGPC_OPEN)          GetProcAddress(hDLL, "EGPCOpen");
        EGPCClose        = (EGPC_CLOSE)         GetProcAddress(hDLL, "EGPCClose");
        EGPCSetBaudRate  = (EGPC_SET_BAUDRATE)  GetProcAddress(hDLL, "EGPCSetBaudRate");
        EGPCGetBaudRate  = (EGPC_GET_BAUDRATE)  GetProcAddress(hDLL, "EGPCGetBaudRate");
        EGPCSetFilter    = (EGPC_SET_FILTER)    GetProcAddress(hDLL, "EGPCSetFilter");
        EGPCClearFilter  = (EGPC_CLEAR_FILTER)  GetProcAddress(hDLL, "EGPCClearFilter");
        EGPCReceive      = (EGPC_RECEIVE)       GetProcAddress(hDLL, "EGPCReceive");
        EGPCSend         = (EGPC_SEND)          GetProcAddress(hDLL, "EGPCSend");
        EGPCReset        = (EGPC_RESET)         GetProcAddress(hDLL, "EGPCReset");
        EGPCShowVer      = (EGPC_SHOW_VER)      GetProcAddress(hDLL, "EGPCShowVer");

        if(!EGPCOpen        || !EGPCClose      || !EGPCSetBaudRate || !EGPCGetBaudRate || !EGPCSetFilter ||
           !EGPCClearFilter || !EGPCReceive    || !EGPCSend        || !EGPCReset       || !EGPCShowVer)
        {
            printf("Get function from %s failed!\n", LIB_FCANBUS);
            exit(1);
        }
    }
} /* END: load_dll() */


/*---------------------------------------------------------------------------------*/
void CALLBACK cb_receive_can_info(long error, CanFrameInfor* message)
{
	int i;
	rxCnt++;

    if(!ini_item.print_rx_raw_data)
    {
        if(rxCnt % ini_item.send_cnt_once == 0)
            printf("[RX count]: %lld\n", rxCnt);
    }
    else
    {
        if(error)
        {
            printf_s("Read error: %Xh\n", error);
            if((error == 0x80050FFF) || (error == 0x8005FFFF))
            {
                errorCnt2++;
            }
            else
            {
                errorCnt1++;
            }

#if 0
            /* ----- EGPCReset() ----- */
            /* If an error occurs, you can reset by EGPCReset() */
            long rtn = EGPCReset(ini_item.comport);
            printf("EGPCReset(rtn=%lx) !\n", rtn);
#endif

        }
        else
        {
            if (message->type == CP_11Bit)
                printf("%lld. [RX] [ID:%03lX] [dlc:%d] ", rxCnt, message->id, message->data_len);
            else
                printf("%lld. [RX] [ID:%08lX] [dlc:%d] ", rxCnt, message->id, message->data_len);

            if (message->rtr)
                printf("[rtr:%d]", message->rtr);
            else
            {
                printf("[rtr:%d] Data: ", message->rtr);
                for (i = 0; i < message->data_len; i++)
                    printf("[%02X] ", message->data[i]);
            }
            printf("\n");
        }
    }
}
