#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <process.h>
#include <windows.h>
#include <time.h>

#include "ini.h"
#include "lib_egpc_b4s1.h"

#define LIB_FCANBUS "lib_egpc_b4s1.dll"

#define  CANBUS_MAX_COMPORT_NUM 4
#define  CANBUS_TEST_BAUDRATE 500000
#define  RECV_TIMEOUT 1 /* [sec] */

#define  OPEN_HINT      "\n ====== Open EGPC_B4S1 ======\n\n"
#define  SCAN_HINT      "\n ====== Scan connecting ======\n\n"
#define  START_HINT     "\n ====== Start test ======\n\n"
#define  TX_RX_PASS     " @ %s -> %s ----- PASS !\n"
#define  TX_RX_FAIL     " @ %s -> %s ----- FAILED !\n"
#define  EXIT_HINT      "\n Press \"Enter\" to exit ..."

HMODULE         hDLL       = NULL;
HANDLE          TRD_send   = NULL;
bool            recv_flag  = false;
bool            recv_error = false;
bool            scan_connect = false;
int             get_connect = 0;
bool            is_new     = false;
bool            tx_rx_success_flag = true;
CanFrameInfor   frame_recv;
char           *log_file;
char            send_comport_str[CANBUS_MAX_COMPORT_STR_LEN + 1];   // for loopback 2 ports
char            recv_comport_str[CANBUS_MAX_COMPORT_STR_LEN + 1];   // for loopback 2 ports
DWORD           id_s       = 0x7FF;
DWORD           id_e       = 0x1FFFFFFF;
unsigned char   data[CANBUS_MAX_COMPORT_NUM][CANBUS_MAX_DATA_SIZE] =
                {
                    {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77},
                    {0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF},
                    {0xFF, 0xEE, 0xDD, 0xCC, 0xBB, 0xAA, 0x99, 0x88},
                    {0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0x00}
                };

EGPC_OPEN           EGPCOpen;
EGPC_CLOSE          EGPCClose;
EGPC_SET_BAUDRATE   EGPCSetBaudRate;
EGPC_GET_BAUDRATE   EGPCGetBaudRate;
EGPC_RECEIVE        EGPCReceive;
EGPC_SEND           EGPCSend;
EGPC_SHOW_VER       EGPCShowVer;

enum
{
    FAIL_LOAD_DLL = 0,
    FAIL_GET_FUNC,
    FAIL_GET_ALL_PORT,
    FAIL_SCAN_CONNECT,
    FAIL_TIME_OUT,
    SUCCESS,
    LOOPBACK,
    FAIL_SEND_RECV,
    NUM
};

const char *log_str[NUM] =
{
  "Load lib_egpc_b4s1.dll failed",
  "Get function from lib_egpc_b4s1.dll failed",
  "Can't find all comport",
  "Scan connecting failed",
  "EMUCReceive time out",
  "EGPC-B4S1 testing successfully"
};

typedef struct
{
    /* [test] */
    int COM_Qty;
    char COM_1th[CANBUS_MAX_COMPORT_STR_LEN + 1];
    char COM_2nd[CANBUS_MAX_COMPORT_STR_LEN + 1];

    /* [setting] */
    DWORD baudrate;
    int   sleep_interval;
    int   test_time;

    /* [data] */
    int data_length;
    bool rtr;

    /* [id] */
    bool id_type;

} INI_ITEM;

typedef struct
{
  int             log_type;
  char           *send_port;
  char           *recv_port;
  DWORD           send_id;
  DWORD           recv_id;
  unsigned char  *send_data;
  unsigned char  *recv_data;
  
} LOG_INFO;

typedef struct
{
    int comport_int;
    char comport_str[CANBUS_MAX_COMPORT_STR_LEN + 1];

    int connect_comport_item;

} COMPORT_ITEM;

INI_ITEM  ini_item;
COMPORT_ITEM comport_item[CANBUS_MAX_COMPORT_NUM];

void CALLBACK cb_receive_can_info_0 (long error, CanFrameInfor* message);
void CALLBACK cb_receive_can_info_1 (long error, CanFrameInfor* message);
void CALLBACK cb_receive_can_info_2 (long error, CanFrameInfor* message);
void CALLBACK cb_receive_can_info_3 (long error, CanFrameInfor* message);
static void cb_by_comport_item(int item_num, long error, CanFrameInfor* message);
static void get_ini_setting (void);
static int  handler_ini (void* user, const char* section, const char* name, const char* value);
static void set_log_file (void);
static void print_config (void);
static int  load_dll (void);
static BOOL end_egpc (DWORD type);
static DWORD WINAPI TRD_send_fx (LPVOID ptr);
static void print_data(CanFrameInfor frame_info, bool is_send, bool is_new);
static int  proc_recv (bool scan_connect, int send_port_item, int id_type);
static void write_log (LOG_INFO log_info);


/*---------------------------------------------------------------------------------*/
int main (int argc, char *argv[])
{
    int                 i, com;
    int                 comNum = 0;
    long                rtn;
    char                ver_info[VER_LEN];
    LOG_INFO            log_info;
    CanFrameInfor       frame_send;
    int                 connect_pair_num = 0;
    char               *exec_cmd;
    SHELLEXECUTEINFO    ShExecInfo[CANBUS_MAX_COMPORT_NUM/2] = {0};

    /* ----- load lib_egpc_b4s1.dll ----- */
    rtn = load_dll();
    if(rtn == 1)
    {
        log_info.log_type = FAIL_LOAD_DLL;
        write_log(log_info);
        goto END;
    }
    else if(rtn == 2)
    {
        log_info.log_type = FAIL_GET_FUNC;
        write_log(log_info);
        goto END;
    }

    EGPCShowVer(ver_info);

    printf("EGPC library version: %s\n", ver_info);

    /* ----- get ini setting ----- */
    get_ini_setting();

    /* ----- enter from 4 port program ----- */
    if(argc > 1)
    {
        ini_item.COM_Qty = 2;

        memcpy(ini_item.COM_1th, argv[1], strlen(argv[1]));
        ini_item.COM_1th[strlen(ini_item.COM_1th)] = '\0';

        memcpy(ini_item.COM_2nd, argv[2], strlen(argv[2]));
        ini_item.COM_2nd[strlen(ini_item.COM_2nd)] = '\0';
    }

    set_log_file();

    print_config();

    if(ini_item.COM_Qty == 4)
    {
        /* ----- scan all CANbus comport & its baudrate ----- */
        printf(OPEN_HINT);

        for(com=1; com<256; com++)
        {
            char com_str[CANBUS_MAX_COMPORT_STR_LEN + 1];

            memset(com_str, 0, strlen(com_str));
            sprintf(com_str, "COM%d", com);

            rtn = EGPCOpen(com_str);
            if(!rtn)
            {
                comport_item[comNum].comport_int = com;
                sprintf(comport_item[comNum].comport_str, "%s", com_str);

                DWORD getBaudRate = 0;
                rtn = EGPCSetBaudRate(comport_item[comNum].comport_str, ini_item.baudrate);
                if(rtn)
                {
                    printf("EGPC Set BaudRate failed (rtn=%lx) !\n", rtn);
                    goto END;
                }
                else
                {
                    rtn = EGPCGetBaudRate(comport_item[comNum].comport_str, &getBaudRate);
                    if(rtn)
                    {
                        printf("EGPC Get BaudRate failed (rtn=%lx) !\n", rtn);
                        goto END;
                    }
                }
                printf("comport: %s; BaudRate: %lu !\n", comport_item[comNum].comport_str, getBaudRate);
                printf("==============================================\n");
                comNum++;
            }

            if(comNum == CANBUS_MAX_COMPORT_NUM)
                break;
        }

        if(comNum != CANBUS_MAX_COMPORT_NUM)
        {
            log_info.log_type = FAIL_GET_ALL_PORT;
            write_log(log_info);
            printf("%s (only %d comports)", log_str[FAIL_GET_ALL_PORT], comNum);
            goto END;
        }


        /* ----- register all comport's receive callback ----- */
        EGPCReceive(comport_item[0].comport_str, cb_receive_can_info_0);
        EGPCReceive(comport_item[1].comport_str, cb_receive_can_info_1);
        EGPCReceive(comport_item[2].comport_str, cb_receive_can_info_2);
        EGPCReceive(comport_item[3].comport_str, cb_receive_can_info_3);

        /* ----- scan all CANbus connecting status ----- */
        printf(SCAN_HINT);
        scan_connect = true;

        memset(&frame_send, 0, sizeof(CanFrameInfor));

        frame_send.type = CP_11Bit;
        frame_send.rtr = DISABLE_RTR;
        frame_send.data_len = 1;

        for(i=0; i<CANBUS_MAX_COMPORT_NUM; i++)
            comport_item[i].connect_comport_item = i; // assume not connect other one

        for(i=0; i<CANBUS_MAX_COMPORT_NUM; i++)
        {
            frame_send.id = i;
            EGPCSend(comport_item[i].comport_str, &frame_send);
            for(int i=0; i<100000; i++); // sleep for AP crash with baudrate less than 50K
        }

        Sleep(RECV_TIMEOUT * 1000);

        for(i=0; i<CANBUS_MAX_COMPORT_NUM; i++)
        {
            if(comport_item[i].connect_comport_item == i)
                printf("%s haven't connect another CANbus\n", comport_item[i].comport_str);
            else
                printf("%s <-> %s\n", comport_item[i].comport_str, comport_item[comport_item[i].connect_comport_item].comport_str);
            EGPCClose(comport_item[i].comport_str);
        }

        if(get_connect != CANBUS_MAX_COMPORT_NUM)
        {
            log_info.log_type = FAIL_SCAN_CONNECT;
            write_log(log_info);
            goto END;
        }

        scan_connect = false;

        /* ---------- start loopback test ---------- */
        connect_pair_num = 0;

        for(i=0; i<CANBUS_MAX_COMPORT_NUM; i++)
        {
            if(comport_item[i].connect_comport_item)
            {
                exec_cmd = malloc(strlen(comport_item[i].comport_str) + strlen(comport_item[comport_item[i].connect_comport_item].comport_str) + 3);
                exec_cmd[0] = '\0';
                strcat(exec_cmd, comport_item[i].comport_str);
                strcat(exec_cmd, " ");
                strcat(exec_cmd, comport_item[comport_item[i].connect_comport_item].comport_str);
                exec_cmd[strlen(exec_cmd)] = '\0';
                printf("Start to test loopback: %s\n", exec_cmd);
                
                ShExecInfo[connect_pair_num].cbSize = sizeof(SHELLEXECUTEINFO);
                ShExecInfo[connect_pair_num].fMask = SEE_MASK_NOCLOSEPROCESS;
                ShExecInfo[connect_pair_num].hwnd = NULL;
                ShExecInfo[connect_pair_num].lpVerb = NULL;
                ShExecInfo[connect_pair_num].lpFile = argv[0];        
                ShExecInfo[connect_pair_num].lpParameters = exec_cmd;   
                ShExecInfo[connect_pair_num].lpDirectory = NULL;
                ShExecInfo[connect_pair_num].nShow = SW_SHOW;
                ShExecInfo[connect_pair_num].hInstApp = NULL; 
                ShellExecuteEx(&ShExecInfo[connect_pair_num]);
                
                free(exec_cmd);
                connect_pair_num++;
            }

            if(connect_pair_num == 2)
                break;
        }
        WaitForSingleObject(ShExecInfo[0].hProcess, INFINITE);
        WaitForSingleObject(ShExecInfo[1].hProcess, INFINITE);
        CloseHandle(ShExecInfo[0].hProcess);
        CloseHandle(ShExecInfo[1].hProcess);
        goto END;
    }
    else if(ini_item.COM_Qty == 2)
    {
        memcpy(send_comport_str, ini_item.COM_1th, strlen(ini_item.COM_1th));
        send_comport_str[strlen(send_comport_str)] = '\0';

        memcpy(recv_comport_str, ini_item.COM_2nd, strlen(ini_item.COM_2nd));
        recv_comport_str[strlen(recv_comport_str)] = '\0';

        EGPCOpen(send_comport_str);
        EGPCOpen(recv_comport_str);
        EGPCSetBaudRate(send_comport_str, ini_item.baudrate);
        EGPCSetBaudRate(recv_comport_str, ini_item.baudrate);

        rtn = EGPCReceive(recv_comport_str, cb_receive_can_info_0);
        if(rtn)
        {
            printf("EGPC Receive failed (rtn=%lx) !\n", rtn);
            exit(1);
        }

        TRD_send = CreateThread(NULL, 0, TRD_send_fx, NULL, 0, NULL);
        SetConsoleCtrlHandler((PHANDLER_ROUTINE) end_egpc, TRUE);
        WaitForSingleObject(TRD_send, INFINITE);

        log_info.log_type = LOOPBACK;
        write_log(log_info);

        if(tx_rx_success_flag)
            printf("EGPC-B4S1 loopback test successfully\n");
        else
            printf("EGPC-B4S1 loopback test failed\n");

        goto END;
    }
    else
    {
        printf("ERROR: setup.ini failed. (illegal COM_Qty)\n");
        goto END;
    }

    return 0;

    END:
        printf(EXIT_HINT);
        getchar();
        return 0;

} /* END: main() */


/*------------------------------------------------------------------------------------*/
static int proc_recv (bool scan_connect, int send_port_item, int id_type)
{
    int              recv_port_item = comport_item[send_port_item].connect_comport_item;
    int              i, dlc;
    bool             fail_flag = false;
    clock_t          period;
    clock_t          start;
    DWORD            check_id;
    unsigned char    check_data[CANBUS_MAX_DATA_SIZE];
    LOG_INFO         log_info;


    dlc = ini_item.data_length;

    if(!scan_connect)
    {
        memcpy(check_data, data[send_port_item], CANBUS_MAX_DATA_SIZE);

        if(id_type == CP_11Bit)  check_id = id_s;
        else                     check_id = id_e;
    }

    period = RECV_TIMEOUT * 1000;
    start  = clock();

    while (1)
    {
        if(recv_flag)
        {
            if(!scan_connect)
            {
                /* check rtr */
                if(frame_recv.rtr != ini_item.rtr)
                {
                    fail_flag = true;
                    break;
                }

                /* check data */
                if(!ini_item.rtr)
                {
                    for(i=0; i<dlc; i++)
                    {
                        if(frame_recv.data[i] != check_data[i])
                        {
                            fail_flag = true;
                            break;
                        }
                    }
                }

                /* check id */
                if(frame_recv.id != check_id)
                    fail_flag = true;

                /* failed */
                if(fail_flag)
                {
                    log_info.log_type  = FAIL_SEND_RECV;
                    log_info.send_port = comport_item[send_port_item].comport_str;
                    log_info.recv_port = comport_item[recv_port_item].comport_str;
                    log_info.send_id   = check_id;
                    log_info.recv_id   = frame_recv.id;
                    log_info.send_data = &check_data[0];
                    log_info.recv_data = &frame_recv.data[0];

                    write_log(log_info);
                    return 1;
                }
            }
            else
            {
                /* check id */
                if(frame_recv.id != comport_item[send_port_item].comport_int)
                    return 1;
            }
            
            recv_flag = false;
            return 0;
        }

        if(clock() > start + period)  /* time out */
        {
            if(!scan_connect)
            {
                log_info.log_type = FAIL_TIME_OUT;
                write_log(log_info);
            }
            return 1;
        }
    }

    return 0;

} /* END: proc_recv() */


/*---------------------------------------------------------------------------------*/
static BOOL end_egpc (DWORD type)
{
    if(type == CTRL_C_EVENT)
    {

        printf("\nExit the program ...\n");

        TerminateThread(TRD_send, 0);

        /* ----- EGPCClose() ----- */
        if(strcmp(send_comport_str, "0"))
            EGPCClose(send_comport_str);

        if(strcmp(recv_comport_str, "0"))
            EGPCClose(recv_comport_str);

        exit(0);
    }

    /* prevent windows compiler warning */
    return TRUE;

} /* END: end_egpc() */


/*---------------------------------------------------------------------------------*/
static DWORD WINAPI TRD_send_fx (LPVOID ptr)
{
    long            rtn;
    CanFrameInfor   frame_send;
    clock_t         start_t;

    memset(&frame_send, 0, sizeof(CanFrameInfor));

    if(ini_item.id_type)
    {
        frame_send.type = CP_29Bit;
        frame_send.id = id_e;
    }
    else
    {
        frame_send.type = CP_11Bit;
        frame_send.id = id_s;
    }

    frame_send.rtr = ini_item.rtr;
    frame_send.data_len = ini_item.data_length;

    memcpy(frame_send.data, data[0], frame_send.data_len);

    start_t = clock();

    while(1)
    {
        /* ----- EGPCSend() ----- */
        is_new = true;
        print_data(frame_send, true, is_new);
        EGPCSend(send_comport_str, &frame_send);

        rtn = proc_recv(false, 0, frame_send.type);
        if(rtn)
        {
            tx_rx_success_flag = false;
            return rtn;
        }
        else
        {
            is_new = false;
            print_data(frame_recv, false, is_new);
        }

        if(ini_item.sleep_interval)
            Sleep(ini_item.sleep_interval);

        if(recv_error)
        {
            tx_rx_success_flag = false;
            return 1;
        }

        if(clock() > (start_t + ini_item.test_time * 1000))  /* time out */
            break;
    }

    tx_rx_success_flag = true;
    return 0;

} /* END: TRD_send_fx() */


/*---------------------------------------------------------------------------------*/
static void get_ini_setting (void)
{
    if(ini_parse("setup.ini", handler_ini, &ini_item) < 0)
    {
        printf("Open setup.ini failed\n");
        exit(1);
    }

} /* END: get_ini_setting() */


/*---------------------------------------------------------------------------------*/
static int handler_ini (void *user, const char *section, const char *name, const char *value)
{
    #define MATCH(s, n) strcmp(section, s) == 0 && strcmp(name, n) == 0

    INI_ITEM *p_ini = (INI_ITEM *)user;

    /* [test] section */
    if(MATCH("test", "COM_Qty"))
    {
        p_ini->COM_Qty = (int) atoi(value);
    }
    else if(MATCH("test", "COM_1th"))
    {
        strcpy(p_ini->COM_1th, value);
        p_ini->COM_1th[strlen(value)] = '\0';
    }
    else if(MATCH("test", "COM_2nd"))
    {
        strcpy(p_ini->COM_2nd, value);
        p_ini->COM_2nd[strlen(value)] = '\0';
    }

    /* [setting] section */
    else if(MATCH("setting", "baudrate"))
    {
        p_ini->baudrate = (DWORD) atoi(value);
    }
    else if(MATCH("setting", "sleep_interval"))
    {
        p_ini->sleep_interval = (int) atoi(value);
    }
    else if(MATCH("setting", "test_time"))
    {
        p_ini->test_time = (int) atoi(value);
    }

    /* [data] section */
    else if(MATCH("data", "data_length"))
    {
        p_ini->data_length = (int) atoi(value);
    }
    else if(MATCH("data", "rtr"))
    {
        p_ini->rtr = (bool) atoi(value);
    }

    /* [id] section */
    else if(MATCH("id", "id_type"))
    {
        p_ini->id_type = (bool) atoi(value);
    }

    /* unknown section/name, error */
    else
    {
        return 0;
    }

    return 1;

} /* END: handler_ini() */


/*---------------------------------------------------------------------------------*/
static void print_config (void)
{
    printf("\nConfig\n");
    printf("----------------------------------\n");
    printf("COM_Qty: %d\n", ini_item.COM_Qty);
    if(ini_item.COM_Qty == 2)
    {
        printf("COM_1th: %s\n", ini_item.COM_1th);
        printf("COM_2nd: %s\n", ini_item.COM_2nd);
    }
    printf("baudrate: %lu\n", ini_item.baudrate);
    printf("sleep_interval: %d\n", ini_item.sleep_interval);
    printf("test_time: %d\n", ini_item.test_time);
    printf("data_length: %d\n", ini_item.data_length);
    printf("rtr: %s\n", ini_item.rtr ? "enable" : "disable");
    printf("id_type: %s\n", ini_item.id_type ? "29Bit" : "11Bit");
    printf("----------------------------------\n\n");

} /* END: print_config() */


/*---------------------------------------------------------------------------------*/
static int load_dll (void)
{
    hDLL = LoadLibrary(LIB_FCANBUS);

    if(hDLL == NULL)
    {
        printf("Load %s failed!(%ld)\n", LIB_FCANBUS, GetLastError());
        return 1;
    }
    else
    {
        EGPCOpen         = (EGPC_OPEN)          GetProcAddress(hDLL, "EGPCOpen");
        EGPCClose        = (EGPC_CLOSE)         GetProcAddress(hDLL, "EGPCClose");
        EGPCSetBaudRate  = (EGPC_SET_BAUDRATE)  GetProcAddress(hDLL, "EGPCSetBaudRate");
        EGPCGetBaudRate  = (EGPC_GET_BAUDRATE)  GetProcAddress(hDLL, "EGPCGetBaudRate");
        EGPCReceive      = (EGPC_RECEIVE)       GetProcAddress(hDLL, "EGPCReceive");
        EGPCSend         = (EGPC_SEND)          GetProcAddress(hDLL, "EGPCSend");
        EGPCShowVer      = (EGPC_SHOW_VER)      GetProcAddress(hDLL, "EGPCShowVer");

        if(!EGPCOpen    || !EGPCClose || !EGPCSetBaudRate || !EGPCGetBaudRate ||
           !EGPCReceive || !EGPCSend  || !EGPCShowVer)
        {
            printf("Get function from %s failed!\n", LIB_FCANBUS);
            return 2;
        }
        return 0;
    }
    return 1;
} /* END: load_dll() */


/*------------------------------------------------------------------------------------*/
static void set_log_file (void)
{
    if(ini_item.COM_Qty == 4)
    {
        log_file = malloc(strlen("result.log"));
        sprintf(log_file, "result.log");
    }
    else
    {
        log_file = malloc(strlen("result.log") + strlen(ini_item.COM_1th) + strlen(ini_item.COM_2nd) + 3);
        log_file[0] = '\0';
        strcat(log_file, "result_");
        strcat(log_file, ini_item.COM_1th);
        strcat(log_file, "_");
        strcat(log_file, ini_item.COM_2nd);
        strcat(log_file, ".log");
        log_file[strlen(log_file)] = '\0';
    }
} /* END: set_log_file() */


/*------------------------------------------------------------------------------------*/
static void write_log (LOG_INFO log_info)
{
    char            time_str[20];
    struct tm      *timeinfo;
    struct timeb    now;
    int             i;
    FILE           *fptr = NULL;

    ftime(&now);
    timeinfo = localtime((time_t*) &now);
    sprintf(time_str, "%d/%02d/%02d %02d:%02d:%02d", (timeinfo->tm_year)+1900,
                                                     (timeinfo->tm_mon)+1,
                                                      timeinfo->tm_mday,
                                                      timeinfo->tm_hour,
                                                      timeinfo->tm_min,
                                                      timeinfo->tm_sec);

    fptr = fopen(log_file, "a");
    if(fptr)
    {
        fprintf(fptr, "======================= %s =======================\n", time_str);

        if(log_info.log_type == LOOPBACK)
        {
            fprintf(fptr, "comport: %s, %s\n", send_comport_str, recv_comport_str);
            fprintf(fptr, "baudrate  = %ld\n", ini_item.baudrate);
            fprintf(fptr, "interval  = %d [ms]\n", ini_item.sleep_interval);
            fprintf(fptr, "test_time = %d [sec]\n", ini_item.test_time);
            if(tx_rx_success_flag)
                fprintf(fptr, "EGPC-B4S1 loopback test successfully\n");
            else
                fprintf(fptr, "EGPC-B4S1 loopback test failed\n");
        }
        else
        {
            if(log_info.log_type != FAIL_SEND_RECV)
                fprintf(fptr, "%s\n", log_str[log_info.log_type]);
            else
            {
                fprintf(fptr, "Send port: %s\n", log_info.send_port);
                fprintf(fptr, "Recv port: %s\n", log_info.recv_port);
                fprintf(fptr, "\n");
                fprintf(fptr, "Send ID  : %lX\n",log_info.send_id);
                fprintf(fptr, "Recv ID  : %lX\n",log_info.recv_id);
                fprintf(fptr, "\n");
                fprintf(fptr, "Send data: ");
                for(i=0; i<CANBUS_MAX_DATA_SIZE; i++)
                    fprintf(fptr, "%02X ", *(log_info.send_data + i));
                fprintf(fptr, "\n");
                fprintf(fptr, "Recv data: ");
                for(i=0; i<CANBUS_MAX_DATA_SIZE; i++)
                    fprintf(fptr, "%02X ", *(log_info.recv_data + i));
                fprintf(fptr, "\n");
            }
        }
        fprintf(fptr, "\n");
        fclose(fptr);
    }
    else
        printf("Create new file failed !\n");


} /* END: write_log() */


/*---------------------------------------------------------------------------------*/
static void print_data(CanFrameInfor frame_info, bool is_send, bool is_new)
{
    static unsigned int round = 1;

    int i;

    if(is_new)
    {
        system("cls");
        printf("Round %u:\n", round);
        printf("===========");
        round++;
    }

    if (is_send)
    {
        printf("\n---------------------------------------------------------\n");
        printf("Send: ");
    }
    else
        printf("Recv: ");

    /* ID */
    if(ini_item.id_type)
        printf("[ID: %08lX] [dlc:%d] ", frame_info.id, frame_info.data_len);
    else
        printf("[ID: %03lX] [dlc:%d] ", frame_info.id, frame_info.data_len);

    /* Data */
    if(frame_info.rtr)
        printf("[rtr:%d]", frame_info.rtr);
    else
    {
        printf("[rtr:%d] Data: ", frame_info.rtr);
        for (i = 0; i<frame_info.data_len; i++)
            printf("[%02X] ", frame_info.data[i]);
    }
    printf("\n");

} /* END: print_data() */


/*---------------------------------------------------------------------------------*/
static void cb_by_comport_item(int item_num, long error, CanFrameInfor* message)
{
	if(!error)
    {
#if 0
		if (message->type == CP_11Bit)
            printf("[RX] [ID:%03lX] [dlc:%d] ", message->id, message->data_len);
		else
            printf("[RX] [ID:%08lX] [dlc:%d] ", message->id, message->data_len);

        if (message->rtr)
            printf("[rtr:%d]", message->rtr);
        else
        {
            printf("[rtr:%d] Data: ", message->rtr);
            for (int i = 0; i < message->data_len; i++)
                printf("[%02X] ", message->data[i]);
        }
        printf("\n");
#endif
        if(scan_connect)
        {
            comport_item[item_num].connect_comport_item = message->id;
            get_connect++;
        }
        else
        {
            memset(&frame_recv, 0, sizeof(frame_recv));
            memcpy(&frame_recv, message, sizeof(CanFrameInfor));
            recv_flag = true;
        }
        
    }
    else
    {
        recv_error = true;
    }
} /* END: cb_by_comport_item() */


/*---------------------------------------------------------------------------------*/
void CALLBACK cb_receive_can_info_0(long error, CanFrameInfor* message)
{
    const int item_num = 0;

    cb_by_comport_item(item_num, error, message);
} /* END: cb_receive_can_info_0() */


/*---------------------------------------------------------------------------------*/
void CALLBACK cb_receive_can_info_1(long error, CanFrameInfor* message)
{
    const int item_num = 1;

    cb_by_comport_item(item_num, error, message);
} /* END: cb_receive_can_info_1() */


/*---------------------------------------------------------------------------------*/
void CALLBACK cb_receive_can_info_2(long error, CanFrameInfor* message)
{
    const int item_num = 2;

    cb_by_comport_item(item_num, error, message);
} /* END: cb_receive_can_info_2() */


/*---------------------------------------------------------------------------------*/
void CALLBACK cb_receive_can_info_3(long error, CanFrameInfor* message)
{
    const int item_num = 3;

    cb_by_comport_item(item_num, error, message);
} /* END: cb_receive_can_info_3() */