#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <time.h>

#include "lib_canopen.h"

#include <conio.h>
#include <windows.h>
#if defined(_M_X64) || defined(__amd64__)
    #define LIB_CANOPEN "lib_canopen_64.dll"
#else
    #define LIB_CANOPEN "lib_canopen_32.dll"
#endif

HMODULE hDLL = NULL;

EGPC_CANOPEN_ENABLE EGPCCANOpenEnable = NULL;
EGPC_CANOPEN_DISABLE EGPCCANOpenDisable = NULL;
EGPC_WRITE_OD EGPCWriteOD = NULL;
EGPC_CANOPEN_TX EGPCCANOpenTx = NULL;
EGPC_CANOPEN_SET_STATE EGPCCANOpenSetState = NULL;
EGPC_CANOPEN_RQST_STATE EGPCCANOpenRqstState = NULL;
EGPC_CANOPEN_CHECK_NODE_STATE EGPCCANOpenCheckNodeState = NULL;
EGPC_CANOPEN_GET_VER EGPCCANOpenGetVer = NULL;

void load_dll(void)
{
    hDLL = LoadLibrary(LIB_CANOPEN);

    if(hDLL == NULL)
    {
        printf("Load %s failed!\n", LIB_CANOPEN);
        exit(1);
    }
    else
    {
        EGPCCANOpenEnable         = (EGPC_CANOPEN_ENABLE)           GetProcAddress(hDLL, "EGPCCANOpenEnable");
        EGPCCANOpenDisable        = (EGPC_CANOPEN_DISABLE)          GetProcAddress(hDLL, "EGPCCANOpenDisable");
        EGPCWriteOD               = (EGPC_WRITE_OD)                 GetProcAddress(hDLL, "EGPCWriteOD");
        EGPCCANOpenTx             = (EGPC_CANOPEN_TX)               GetProcAddress(hDLL, "EGPCCANOpenTx");
        EGPCCANOpenSetState       = (EGPC_CANOPEN_SET_STATE)        GetProcAddress(hDLL, "EGPCCANOpenSetState");
        EGPCCANOpenRqstState      = (EGPC_CANOPEN_RQST_STATE)       GetProcAddress(hDLL, "EGPCCANOpenRqstState");
        EGPCCANOpenCheckNodeState = (EGPC_CANOPEN_CHECK_NODE_STATE) GetProcAddress(hDLL, "EGPCCANOpenCheckNodeState");
        EGPCCANOpenGetVer         = (EGPC_CANOPEN_GET_VER)          GetProcAddress(hDLL, "EGPCCANOpenGetVer");

        if(!EGPCCANOpenEnable || !EGPCCANOpenDisable || !EGPCWriteOD || !EGPCCANOpenTx || !EGPCCANOpenSetState ||
           !EGPCCANOpenRqstState || !EGPCCANOpenCheckNodeState || !EGPCCANOpenGetVer)
        {
            printf("Get function from %s failed!\n", LIB_CANOPEN);
            exit(1);
        }
    }
}

void cb_handle_sdo_abort(uint8_t p, uint8_t n, uint16_t index, uint8_t subindex, uint8_t cs, uint32_t abort_code);
void cb_handle_sdo(uint8_t p, uint8_t n, uint16_t index, uint8_t subindex, uint8_t *buf, uint8_t buf_len);
void cb_cob_handle(uint8_t port, uint8_t fcode, uint8_t src, uint8_t *buf, uint8_t buf_len);
void cb_update_handle(int port);
void cb_set_state(uint8_t p, uint8_t state);
void cb_time(uint8_t p, uint32_t milliseconds, uint16_t days);
void cb_sync(uint8_t p, uint8_t counter);
void cb_init_device(uint8_t p);
void cb_can_rx(uint8_t p, uint32_t id, uint8_t *buf, uint8_t buf_len);
void cb_can_tx(uint8_t p, uint32_t id, uint8_t *buf, uint8_t buf_len);

/*----------------------------------------------------------------------*/
int main (int argc, char *argv[])
{
    long   rtn = -1;
    char  *setup_ini = "setup_slave.ini";
    uint8_t slave_port = 0;
    char version[VER_LEN];
    CANOPEN_CB_INFO cb_info;

    load_dll();

    EGPCCANOpenGetVer(version);
    printf("EGPCCANopen ver.%s\n\n", version);

    cb_info.canopen_handle_sdo_abort_cb = cb_handle_sdo_abort;
    cb_info.canopen_handle_sdo_cb = cb_handle_sdo;
    cb_info.canopen_cob_cb = cb_cob_handle;
    cb_info.canopen_update_cb = cb_update_handle;
    cb_info.canopen_set_state_cb = cb_set_state;
    cb_info.canopen_time_cb = cb_time;
    cb_info.canopen_sync_cb = cb_sync;
    cb_info.canopen_init_device_cb = cb_init_device;
    cb_info.canopen_can_rx_cb = cb_can_rx;
    cb_info.canopen_can_tx_cb = cb_can_tx;

    rtn = EGPCCANOpenEnable(setup_ini, cb_info);
    if(rtn)
    {
        printf("EGPCCANOpenEnable() failed! (rtn=%lx)\n", rtn);
        return rtn;
    }
    else
    {
        printf("EGPCCANOpenEnable() successfully!\n");
    }

    Sleep(1000);

    for(;;)
    {
        uint16_t ix = 0x3000;
        uint8_t sx = 0x00;

        /* for writeOD */
        time_t rawtime;
        struct tm *timeinfo;
        uint8_t buf[4];
        uint8_t bl = 3;

        time(&rawtime);
        timeinfo = localtime(&rawtime);
        asctime(timeinfo);
        buf[0] = timeinfo->tm_sec;
        buf[1] = timeinfo->tm_min;
        buf[2] = timeinfo->tm_hour;
        printf("\nslave EGPCWriteOD(slave_port=%d, ix=%04X, sx=%02X, buf, bl)\n", slave_port, ix, sx);
        EGPCWriteOD(slave_port, ix, sx, buf, bl);

        Sleep(1000);

        if(kbhit() != 0)
            break;
    }

    printf("EGPCCANOpenDisable()\n\n");
    EGPCCANOpenDisable();

    return 0;
}


/* callback function */
/*----------------------------------------------------------------------*/
void cb_handle_sdo_abort(uint8_t p, uint8_t n, uint16_t index, uint8_t subindex, uint8_t cs, uint32_t abort_code)
{

}

/*----------------------------------------------------------------------*/
void cb_handle_sdo(uint8_t p, uint8_t n, uint16_t index, uint8_t subindex, uint8_t *buf, uint8_t buf_len)
{

}

/*----------------------------------------------------------------------*/
void cb_cob_handle(uint8_t port, uint8_t fcode, uint8_t src, uint8_t *buf, uint8_t buf_len)
{

}

/*----------------------------------------------------------------------*/
void cb_update_handle(int port)
{

}

/*----------------------------------------------------------------------*/
void cb_set_state(uint8_t p, uint8_t state)
{

}

/*----------------------------------------------------------------------*/
void cb_time(uint8_t p, uint32_t milliseconds, uint16_t days)
{

}

/*----------------------------------------------------------------------*/
void cb_sync(uint8_t p, uint8_t counter)
{

}

/*----------------------------------------------------------------------*/
void cb_init_device(uint8_t p)
{

}

/*----------------------------------------------------------------------*/
void cb_can_rx(uint8_t p, uint32_t id, uint8_t *buf, uint8_t buf_len)
{
    if(p == 0)
    {
        printf("path:RX\tid:%03X\tdlc:%d\tdata:", id, buf_len);
        for(uint8_t i=0; i<buf_len; i++)
            printf("[%02X] ", buf[i]);
        printf("\n");
        if(id == 0x586)
        {
            int hour = buf[6];
            int min = buf[5];
            int sec = buf[4];

            printf("current time: %02d:%02d:%02d (from cb_can_rx())\n\n", hour, min, sec);
        }
    }
}

/*----------------------------------------------------------------------*/
void cb_can_tx(uint8_t p, uint32_t id, uint8_t *buf, uint8_t buf_len)
{
    if(p == 0)
    {
        printf("path:TX\tid:%03X\tdlc:%d\tdata:", id, buf_len);
        for(uint8_t i=0; i<buf_len; i++)
            printf("[%02X] ", buf[i]);
        printf("\n");
    }
}

/*----------------------------------------------------------------------*/
/* END: callback function */

