#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include "ini.h"

#include <sys/timeb.h>
#include <net/if.h>
#include "loopback.h"

INI_ITEM  ini_item;

/* pthread */
pthread_t    TRD_send[2] = {0, 0};
pthread_t    TRD_recv = 0;

bool tx_rx_mutex[2] = {false, false};
int tx_cnt[2] = {0, 0};
int rx_cnt[2] = {0, 0};
int pair = 0b00;

static void *TRD_send_fx (void *ptr);
static void *TRD_recv_fx (void *ptr);
static void  get_ini_setting (void);
static void  print_config (void);
static int   handler_ini (void *user, const char *section, const char *name, const char *value);
static void  write_log (char *log_str);

int candumpMain(int argc, char **argv);
int cansendMain (int argc, char *argv[]);

extern struct timeval diff;
extern char devname[MAXIFNAMES][IFNAMSIZ+1];

static void *TRD_send_fx (void *ptr)
{
    /* ./loopback_send ${pair_1_device_1} ${id_type} ${rtr} ${data_length} ${sleep_interval} ${test_time} */
	char  cmd0[16] = "./loopback_send";
	char *cmd1     = (char*) ptr;
	char  cmd2[2]  = "";
	char  cmd3[2]  = "";
	char  cmd4[2]  = "";
	char  cmd5[16] = "";
    char  cmd6[16] = "";

    sprintf(cmd2, "%d", ini_item.id_type);
    sprintf(cmd3, "%d", ini_item.rtr);
    sprintf(cmd4, "%d", ini_item.data_length);
    sprintf(cmd5, "%d", ini_item.sleep_interval);
    sprintf(cmd6, "%d", ini_item.test_time);

	char *cmd_ptr[7] = {cmd0, cmd1, cmd2, cmd3, cmd4, cmd5, cmd6};
	char **cmd_ptr_ptr = cmd_ptr;

    usleep(100 * 1000);

	int __attribute__ ((unused)) result = cansendMain(sizeof(cmd_ptr)/sizeof(cmd_ptr[0]), cmd_ptr_ptr);
    // printf("=====>> result:%d [%s][%d]\n", result, __FILE__, __LINE__);
    pthread_exit(0);
    return NULL;
}

static void *TRD_recv_fx (void *ptr)
{
    char **cmd_ptr_ptr = NULL;
	int  *pair = (int*) ptr;  // 0b01:pair1 0b10:pair2 0b11:all
    int   argc = 0;

    /* ./loopback_dump ${pair_1_device_2} -t z -T ${dump_time} */
	char  cmd0[16] = "./loopback_dump";
    char *cmd11    = ini_item.pair_1_device_2;
    char *cmd12    = ini_item.pair_2_device_2;
	char  cmd2[3]  = "-t";
	char  cmd3[2]  = "z";
	char  cmd4[3]  = "-T";
	char  cmd5[16] = "";

    if(ini_item.sleep_interval >= 1000)
        sprintf(cmd5, "%d", ini_item.test_time/1000 + 1000);
    else
        sprintf(cmd5, "%d", 1000);

    char *cmd_ptr1[6] = {cmd0, cmd11, cmd2, cmd3, cmd4, cmd5};
    char *cmd_ptr2[6] = {cmd0, cmd12, cmd2, cmd3, cmd4, cmd5};
	char *cmd_ptr3[7] = {cmd0, cmd11, cmd12, cmd2, cmd3, cmd4, cmd5};

    if(*pair == 0b01)
    {
	    cmd_ptr_ptr = cmd_ptr1;
        argc = sizeof(cmd_ptr1)/sizeof(cmd_ptr1[0]);
    }
    else if(*pair == 0b10)
    {
	    cmd_ptr_ptr = cmd_ptr2;
        argc = sizeof(cmd_ptr2)/sizeof(cmd_ptr2[0]);
    }
    else if(*pair == 0b11)
    {
	    cmd_ptr_ptr = cmd_ptr3;
        argc = sizeof(cmd_ptr3)/sizeof(cmd_ptr3[0]);
    }

    int __attribute__ ((unused)) result = candumpMain(argc, cmd_ptr_ptr);
    // printf("=====>> result:%d [%s][%d]\n", result, __FILE__, __LINE__);
    pthread_exit(0);
    return NULL;
}

int main()
{
    char log_str[256];
    char str_tmp[32];
    bool isInverse = false;

    get_ini_setting();
    print_config();

    if(strcmp(ini_item.pair_1_device_1, "0") && strcmp(ini_item.pair_1_device_2, "0"))
        pair += 0b01;

    if(strcmp(ini_item.pair_2_device_1, "0") && strcmp(ini_item.pair_2_device_2, "0"))
        pair += 0b10;

    if(pair)
        pthread_create(&TRD_recv, NULL, TRD_recv_fx, &pair);

    if(pair & 0b01)
        pthread_create(&TRD_send[0], NULL, TRD_send_fx, ini_item.pair_1_device_1);
    
    if(pair & 0b10)
        pthread_create(&TRD_send[1], NULL, TRD_send_fx, ini_item.pair_2_device_1);

    if(pair)
        pthread_join(TRD_recv, NULL);

    if(!strcmp(devname[0], ini_item.pair_1_device_2))
        isInverse = false;
    else if(!strcmp(devname[0], ini_item.pair_2_device_2))
        isInverse = true;
    else
    {
        printf("Please check setup.ini & comport connection status.\n");
        return 1;
    }

    memset(log_str, 0, strlen(log_str));
    log_str[0] = '\0';

    if(pair & 0b01)
    {
        int *result_rx = NULL;

        if(isInverse)
            result_rx = &rx_cnt[1];
        else
            result_rx = &rx_cnt[0];

        strcat(log_str, "Pair1:\n");
        memset(str_tmp, 0, strlen(str_tmp));
        sprintf(str_tmp, "tx_cnt:%d, ", tx_cnt[0]);
        strcat(log_str, str_tmp);
        memset(str_tmp, 0, strlen(str_tmp));
        sprintf(str_tmp, "rx_cnt:%d\n", *result_rx);
        strcat(log_str, str_tmp);
        
        strcat(log_str, "Result: ");
        if(tx_cnt[0] == *result_rx)
            strcat(log_str, "PASS!\n");
        else
            strcat(log_str, "FAIL!\n");
    }

    if(pair & 0b10)
    {
        int *result_rx = NULL;

        if(isInverse)
            result_rx = &rx_cnt[0];
        else
            result_rx = &rx_cnt[1];

        strcat(log_str, "Pair2:\n");
        memset(str_tmp, 0, strlen(str_tmp));
        sprintf(str_tmp, "tx_cnt:%d, ", tx_cnt[1]);
        strcat(log_str, str_tmp);
        memset(str_tmp, 0, strlen(str_tmp));
        sprintf(str_tmp, "rx_cnt:%d\n", *result_rx);
        strcat(log_str, str_tmp);

        strcat(log_str, "Result: ");
        if(tx_cnt[1] == *result_rx)
            strcat(log_str, "PASS!\n");
        else
            strcat(log_str, "FAIL!\n");
    }

    if(pair)
    {
        printf("%s", log_str);
        write_log(log_str);
    }

	return 0;
}


/*---------------------------------------------------------------------------------*/
static void get_ini_setting (void)
{
    if(ini_parse("setup.ini", handler_ini, &ini_item) < 0)
    {
        printf("Open setup.ini failed\n");
        exit(1);
    }

} /* END: get_ini_setting() */


/*---------------------------------------------------------------------------------*/
static int handler_ini (void *user, const char *section, const char *name, const char *value)
{
    #define MATCH(s, n) strcmp(section, s) == 0 && strcmp(name, n) == 0

    INI_ITEM *p_ini = (INI_ITEM *)user;

    /* [setting] section */
    if(MATCH("setting", "pair_1_device_1"))
    {
        strcpy(p_ini->pair_1_device_1, value);
        p_ini->pair_1_device_1[strlen(value)] = '\0';
    }
    else if(MATCH("setting", "pair_1_device_2"))
    {
        strcpy(p_ini->pair_1_device_2, value);
        p_ini->pair_1_device_2[strlen(value)] = '\0';
    }
    else if(MATCH("setting", "pair_2_device_1"))
    {
        strcpy(p_ini->pair_2_device_1, value);
        p_ini->pair_2_device_1[strlen(value)] = '\0';
    }
    else if(MATCH("setting", "pair_2_device_2"))
    {
        strcpy(p_ini->pair_2_device_2, value);
        p_ini->pair_2_device_2[strlen(value)] = '\0';
    }
    else if(MATCH("setting", "sleep_interval"))
    {
        p_ini->sleep_interval = (int) atoi(value);
    }
    else if(MATCH("setting", "test_time"))
    {
        p_ini->test_time = (int) atoi(value);
    }
    else if(MATCH("setting", "monitor"))
    {
        p_ini->monitor = (int) atoi(value);
    }

    /* [data] section */
    else if(MATCH("data", "data_length"))
    {
        p_ini->data_length = (int) atoi(value);
    }
    else if(MATCH("data", "rtr"))
    {
        p_ini->rtr = (bool) atoi(value);
    }

    /* [id] section */
    else if(MATCH("id", "id_type"))
    {
        p_ini->id_type = (bool) atoi(value);
    }

    /* unknown section/name, error */
    else
    {
        return 0;
    }

    return 1;

} /* END: handler_ini() */


/*---------------------------------------------------------------------------------*/
static void print_config (void)
{
    printf("\nConfig\n");
    printf("----------------------------------\n");
    printf("pair_1_device_1: %s\n", ini_item.pair_1_device_1);
    printf("pair_1_device_2: %s\n", ini_item.pair_1_device_2);
    printf("pair_2_device_1: %s\n", ini_item.pair_2_device_1);
    printf("pair_2_device_2: %s\n", ini_item.pair_2_device_2);
    printf("sleep_interval: %d\n", ini_item.sleep_interval);
    printf("test_time: %d\n", ini_item.test_time);
    printf("data_length: %d\n", ini_item.data_length);
    printf("rtr: %s\n", ini_item.rtr ? "enable" : "disable");
    printf("id_type: %s\n", ini_item.id_type ? "29Bit" : "11Bit");
    printf("----------------------------------\n\n");

} /* END: print_config() */


/*---------------------------------------------------------------------------------*/
static void  write_log (char *log_str)
{
    char            time_str[20];
    struct tm      *timeinfo;
    struct timeb    now;
    FILE           *fptr = NULL;

    ftime(&now);
    timeinfo = localtime((time_t*) &now);
    sprintf(time_str, "%d/%02d/%02d %02d:%02d:%02d", (timeinfo->tm_year)+1900,
                                                     (timeinfo->tm_mon)+1,
                                                      timeinfo->tm_mday,
                                                      timeinfo->tm_hour,
                                                      timeinfo->tm_min,
                                                      timeinfo->tm_sec);

    fptr = fopen("result.log", "a");
    if(fptr)
    {
        fprintf(fptr, "======================= %s =======================\n", time_str);
#if 0
        fprintf(fptr, "\nConfig\n");
        fprintf(fptr, "----------------------------------\n");
        fprintf(fptr, "pair_1_device_1: %s\n", ini_item.pair_1_device_1);
        fprintf(fptr, "pair_1_device_2: %s\n", ini_item.pair_1_device_2);
        fprintf(fptr, "pair_2_device_1: %s\n", ini_item.pair_2_device_1);
        fprintf(fptr, "pair_2_device_2: %s\n", ini_item.pair_2_device_2);
        fprintf(fptr, "sleep_interval: %d\n", ini_item.sleep_interval);
        fprintf(fptr, "test_time: %d\n", ini_item.test_time);
        fprintf(fptr, "data_length: %d\n", ini_item.data_length);
        fprintf(fptr, "rtr: %s\n", ini_item.rtr ? "enable" : "disable");
        fprintf(fptr, "id_type: %s\n", ini_item.id_type ? "29Bit" : "11Bit");
        fprintf(fptr, "----------------------------------\n\n");
#endif
        fprintf(fptr, "Loopback test\n");
        fprintf(fptr, "----------------------------------\n");
        fprintf(fptr, "Interval: %03ld.%06ld (sec)\n", diff.tv_sec, diff.tv_usec);
        fprintf(fptr, "%s", log_str);
        fprintf(fptr, "----------------------------------\n\n");

        fclose(fptr);
    }
    else
        printf("Create new file failed !\n");
} /* END: write_log() */
