#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <linux/can.h>
#include <linux/can/raw.h>

#include "lib.h"

#include <sys/timeb.h>
#include <time.h>

#include "loopback.h"

static char *get_now_time (time_t *now_sec);
static char __attribute__ ((unused)) *cal_itvl (char *start_time, char *end_time);
static void split (char **arr, char *str, const char *del);
static void get_config (char *argv[], CFG_ITEM *cfg_item);
static void fill_can_frame(char *can_frame, CFG_ITEM *cfg_item);

char *id_s = "1FF";
char *id_e = "1FFFFFFF";
char data[CANBUS_MAX_DATA_SIZE][3] = {"11", "22", "33", "44", "55", "66", "77", "88"};
unsigned int hex_id_s = 0x1FF;
unsigned int hex_id_e = 0x1FFFFFFF;
unsigned char hex_data[CANBUS_MAX_DATA_SIZE] = {0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88};

extern volatile int running;
extern bool tx_rx_mutex[2];
extern int tx_cnt[2];
extern int rx_cnt[2];
extern int pair;
extern INI_ITEM ini_item;

bool check_inverse (void);

int cansendMain (int argc, char *argv[])
{
#if 0
    printf("argc:%d\n", argc);
	for(int i=0; i<argc; i++)
	{
		printf("argv[%d]:%s\n", i, argv[i]);
	}
	// return 0;
#endif
	int s; /* can raw socket */ 
	int required_mtu;
	int mtu;
	int enable_canfd = 1;
	struct sockaddr_can addr;
	struct canfd_frame frame;
	struct ifreq ifr;
    char __attribute__ ((unused)) *start_time = NULL;
    char __attribute__ ((unused)) *end_time = NULL;
    char can_frame[32];
    CFG_ITEM cfg_item;
    bool *mutex_ptr = NULL;
    int *tx_cnt_ptr = NULL;
    int *rx_cnt_ptr = NULL;

	/* check command line options */
	if (argc != 7) {
		// fprintf(stderr, "Usage: %s <device> <can_frame> <count>.\n", argv[0]);
        fprintf(stderr, "Usage: %s <device> <id_type> <rtr> <data_length> <sleep_interval> <test_time>.\n", argv[0]);
		return 1;
	}

    get_config(argv, &cfg_item);

    memset(can_frame, 0, sizeof(can_frame));
    fill_can_frame(can_frame, &cfg_item);

    printf("send can frame: %s\n\n", can_frame);

	/* parse CAN frame */
	// required_mtu = parse_canframe(argv[2], &frame);
    required_mtu = parse_canframe(can_frame, &frame);
	if (!required_mtu){
		fprintf(stderr, "\nWrong CAN-frame format! Try:\n\n");
		fprintf(stderr, "    <can_id>#{data}            for 'classic' CAN 2.0 data frames\n");
		fprintf(stderr, "    <can_id>#R{len}            for 'classic' CAN 2.0 RTR frames\n");
		fprintf(stderr, "    <can_id>##<flags>{data}    for CAN FD frames\n\n");
		fprintf(stderr, "<can_id> can have 3 (SFF) or 8 (EFF) hex chars\n");
		fprintf(stderr, "{data} has 0..8 (0..64 CAN FD) ASCII hex-values (optionally");
		fprintf(stderr, " separated by '.')\n");
		fprintf(stderr, "{len} is an optional 0..8 value as RTR frames can contain a");
		fprintf(stderr, " valid dlc field\n");
		fprintf(stderr, "<flags> a single ASCII Hex value (0 .. F) which defines");
		fprintf(stderr, " canfd_frame.flags\n\n");
		fprintf(stderr, "e.g. 5A1#11.2233.44556677.88 / 123#DEADBEEF / 5AA# / ");
		fprintf(stderr, "123##1 / 213##311223344\n     1F334455#1122334455667788 / ");
		fprintf(stderr, "123#R / 00000123#R3\n\n");
		return 1;
	}

	/* open socket */
	if ((s = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
		perror("socket");
		return 1;
	}

    // strncpy(ifr.ifr_name, argv[1], IFNAMSIZ - 1);
	strncpy(ifr.ifr_name, cfg_item.send_CAN, IFNAMSIZ - 1);
	ifr.ifr_name[IFNAMSIZ - 1] = '\0';
	ifr.ifr_ifindex = if_nametoindex(ifr.ifr_name);
	if (!ifr.ifr_ifindex) {
		perror("if_nametoindex");
		return 1;
	}

    if(!strcmp(ifr.ifr_name, ini_item.pair_1_device_1))
    {
        mutex_ptr = &tx_rx_mutex[0];
        tx_cnt_ptr = &tx_cnt[0];
        rx_cnt_ptr = &rx_cnt[0];
    }
    else if(!strcmp(ifr.ifr_name, ini_item.pair_2_device_1))
    {
        mutex_ptr = &tx_rx_mutex[1];
        tx_cnt_ptr = &tx_cnt[1];
        if(pair >= 0b11)
            rx_cnt_ptr = &rx_cnt[1];
        else
            rx_cnt_ptr = &rx_cnt[0];
    }
    else
        return -1;

    if(!mutex_ptr || !tx_cnt_ptr || !rx_cnt_ptr)
        return -2;

	memset(&addr, 0, sizeof(addr));
	addr.can_family = AF_CAN;
	addr.can_ifindex = ifr.ifr_ifindex;

	if (required_mtu > CAN_MTU) {

		/* check if the frame fits into the CAN netdevice */
		if (ioctl(s, SIOCGIFMTU, &ifr) < 0) {
			perror("SIOCGIFMTU");
			return 1;
		}
		mtu = ifr.ifr_mtu;

		if (mtu != CANFD_MTU) {
			printf("CAN interface is not CAN FD capable - sorry.\n");
			return 1;
		}

		/* interface is ok - try to switch the socket into CAN FD mode */
		if (setsockopt(s, SOL_CAN_RAW, CAN_RAW_FD_FRAMES,
			       &enable_canfd, sizeof(enable_canfd))){
			printf("error when enabling CAN FD support\n");
			return 1;
		}

		/* ensure discrete CAN FD length values 0..8, 12, 16, 20, 24, 32, 64 */
		frame.len = can_dlc2len(can_len2dlc(frame.len));
	}

	/* disable default receive filter on this RAW socket */
	/* This is obsolete as we do not read from the socket at all, but for */
	/* this reason we can remove the receive list in the Kernel to save a */
	/* little (really a very little!) CPU usage.                          */
	setsockopt(s, SOL_CAN_RAW, CAN_RAW_FILTER, NULL, 0);

	if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		perror("bind");
		return 1;
	}

	/* send frame */
    // int total = atoi(argv[3]);
    time_t start_sec, now_sec, mutex_sec;
    time_t check_cnt_sec = 0;

    start_time = get_now_time(&start_sec);

    // while(count++ < total)
    while(running)
    {
        *mutex_ptr = true;

        if (write(s, &frame, required_mtu) != required_mtu) {
            *mutex_ptr = false;
            perror("write");
            end_time = get_now_time(&now_sec);
#if 0
            printf("send failed! (tx: %d count)\n", *tx_cnt_ptr);
            printf("Start time: %s\n", start_time);
            printf("End time:   %s\n", end_time);
            printf("Interval:   %s\n", cal_itvl(start_time, end_time));
#endif
            return 1;
        }

        (*tx_cnt_ptr)++;

        if(cfg_item.sleep_interval)
            usleep(cfg_item.sleep_interval);

        get_now_time(&now_sec);

        if(!cfg_item.test_time)  /* test one time */
            break;

        if(now_sec > (start_sec + cfg_item.test_time))  /* time out */
            break;

        if(*mutex_ptr)
            get_now_time(&mutex_sec);

        while(*mutex_ptr == true)
        {
            get_now_time(&now_sec);

            if(now_sec > (mutex_sec + cfg_item.sleep_interval/(1000*1000) + 3))  /* time out (sleep_interval + 100ms) */
            {
                running = 0;
                close(s);
                return -3;
            }
        }

        if(!check_cnt_sec)
        {
            if(pair >= 0b11 && check_inverse())
            {
                if(rx_cnt_ptr == &rx_cnt[0])
                    rx_cnt_ptr = &rx_cnt[1];
                else
                    rx_cnt_ptr = &rx_cnt[0];
            }

            get_now_time(&check_cnt_sec);
        }

        if(now_sec > (check_cnt_sec + 5))
        {
            if(*tx_cnt_ptr != *rx_cnt_ptr)
            {
                running = 0;
                close(s);
                return -4;
            }
            get_now_time(&check_cnt_sec);
        }
    }
	close(s);

#if 0
    end_time = get_now_time(&now_sec);
    printf("send successfully! (tx: %d count)\n", *tx_cnt_ptr);
    printf("Start time: %s\n", start_time);
    printf("End time:   %s\n", end_time);
    printf("Interval:   %s\n", cal_itvl(start_time, end_time));

    printf("\n Press \"Enter\" to exit ...");
	getchar();
#endif

	return 0;
} /* END: main() */


/*------------------------------------------------------------------*/
static char *get_now_time (time_t *now_sec)
{
    #define  TIME_CHAR_NUM  13

    char          time_str[TIME_CHAR_NUM];
    struct tm    *timeinfo;
    struct timeb  now;

    ftime(&now);
    timeinfo = localtime((time_t*) &now);
    sprintf(time_str, "%02d:%02d:%02d:%03d", timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec, now.millitm);

    *now_sec = now.time;

    return strdup(time_str);
} /* END: get_now_time() */


/*------------------------------------------------------------------*/
static char *cal_itvl (char *start_time, char *end_time)
{
    int i;
    int j;
    const char *del = ":";
    char rtn_str[12 + 1];

    /* 4: hour/min/sec/ms */
    char *start_arr[4];
    char *end_arr[4];
    int start_int[4];
    int end_int[4];
    int itvl_int[4] = {0};

    split(start_arr, start_time, del);
    split(end_arr, end_time, del);

    i = 0;
    j = 0;
    while(i<4 && j<4)
    {
        start_int[i++] = atoi(*(start_arr+j++));
    }

    i = 0;
    j = 0;
    while(i<4 && j<4)
    {
        end_int[i++] = atoi(*(end_arr+j++));
    }

    /* now we get hour/min/sec/ms of both start time and end time */
    if(end_int[0] >= start_int[0])
    {
        /* ms */
        if(end_int[3] >= start_int[3])
        {
            itvl_int[3] = end_int[3] - start_int[3];
        }
        else
        {
            itvl_int[3] = end_int[3] + 1000 - start_int[3];
            end_int[2] = end_int[2] - 1;
        }

        /* sec */
        if(end_int[2] >= start_int[2])
        {
            itvl_int[2] = end_int[2] - start_int[2];
        }
        else
        {
            itvl_int[2] = end_int[2] + 60 - start_int[2];
            end_int[1] = end_int[1] - 1;
        }

        /* min */
        if(end_int[1] >= start_int[1])
        {
            itvl_int[1] = end_int[1] - start_int[1];
        }
        else
        {
            itvl_int[1] = end_int[1] + 60 - start_int[1];
            end_int[0] = end_int[0] - 1;
        }
    }
    /* the next day */
    else
    {
        int total_start_int[4] = {0};
        int total_end_int[4] = {0};

        /* get start total itvl */
        if(start_int[3] != 0)
        {
            total_start_int[3] = 1000 - start_int[3];
            start_int[2]++;
        }
        else
        {
            total_start_int[3] = 0;
        }

        if(start_int[2] != 0)
        {
            total_start_int[2] = 60 - start_int[2];
            start_int[1]++;
        }
        else
        {
            total_start_int[2] = 0;
        }

        if(start_int[1] != 0)
        {
            total_start_int[1] = 60 - start_int[1];
            start_int[0]++;
        }
        else
        {
            total_start_int[1] = 0;
        }

        total_start_int[0] = 24 - start_int[0];


        /* get end total itvl */
        int i;
        for(i=0; i<4; i++)
        {
            total_end_int[i] = end_int[i];
        }

        /* get all itvl */
        for(i=0; i<4; i++)
        {
            itvl_int[i] = total_start_int[i] + total_end_int[i];
        }

        if(itvl_int[3] >= 1000)
        {
            itvl_int[3] = itvl_int[3] - 1000;
            itvl_int[2]++;
        }

        if(itvl_int[2] >= 60)
        {
            itvl_int[2] = itvl_int[2] - 60;
            itvl_int[1]++;
        }

        if(itvl_int[1] >= 60)
        {
            itvl_int[1] = itvl_int[1] - 60;
            itvl_int[0]++;
        }
    }

    sprintf(rtn_str, "%02d:%02d:%02d:%03d", itvl_int[0], itvl_int[1], itvl_int[2], itvl_int[3]);
    return strdup(rtn_str);
} /* END: cal_itvl() */


/*----------------------------------------------------------------------*/
static void split (char **arr, char *str, const char *del)
{
    char *s = strtok(str, del);

    while(s != NULL)
    {
        *arr++ = s;
        s = strtok(NULL, del);
    }
} /* END: split() */


/*---------------------------------------------------------------------------------*/
static void get_config (char *argv[], CFG_ITEM *cfg_item)
{
    strcpy(cfg_item->send_CAN, argv[1]);
    cfg_item->send_CAN[strlen(cfg_item->send_CAN)] = '\0';

    cfg_item->id_type = (strcmp(argv[2], "0")) ? true : false;
    cfg_item->rtr = (strcmp(argv[3], "0")) ? true : false;
    cfg_item->data_length = (int) atoi(argv[4]);
    cfg_item->sleep_interval = (int) atoi(argv[5]);
    cfg_item->test_time = (int) atoi(argv[6]);

    printf("\nConfig\n");
    printf("----------------------------------\n");
    printf("send_CAN: %s\n", cfg_item->send_CAN);
    printf("id_type: %s\n", cfg_item->id_type ? "29Bit" : "11Bit");
    printf("rtr: %s\n", cfg_item->rtr ? "Enable" : "Disable");
    printf("data_length: %d\n", cfg_item->data_length);
    printf("sleep_interval: %d\n", cfg_item->sleep_interval);
    printf("test_time: %d\n", cfg_item->test_time);
    printf("----------------------------------\n\n");

} /* END: print_config() */


/*---------------------------------------------------------------------------------*/
static void fill_can_frame(char *can_frame, CFG_ITEM *cfg_item)
{
    int i;

    if(cfg_item->id_type)
        strcat(can_frame, id_e);
    else
        strcat(can_frame, id_s);

    strcat(can_frame, "#");

    if(cfg_item->rtr)
    {
        strcat(can_frame, "R");
        char dlc_str[2];
        sprintf(dlc_str, "%d", cfg_item->data_length);
        dlc_str[1] = '\0';
        strcat(can_frame, dlc_str);
    }
    else
    {
        for(i=0; i<cfg_item->data_length; i++)
            strcat(can_frame, data[i]);
    }

    can_frame[strlen(can_frame)] = '\0';

} /* END: fill_can_frame() */
