#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <time.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <net/if.h>

#include "lib_canopen.h"
#include "ini.h"

/*----------------------------------------------------------------------*/
#define FILE_PATH_MAX   1024

/*----------------------------------------------------------------------*/
extern int kbhit(void);

/*----------------------------------------------------------------------*/
void cb_handle_sdo_abort(uint8_t p, uint8_t n, uint16_t index, uint8_t subindex, uint8_t cs, uint32_t abort_code);
void cb_handle_sdo      (uint8_t p, uint8_t n, uint16_t index, uint8_t subindex, uint8_t *buf, uint8_t buf_len);
void cb_cob_handle      (uint8_t port, uint8_t fcode, uint8_t src, uint8_t *buf, uint8_t buf_len);
void cb_update_handle   (int port);
void cb_set_state       (uint8_t p, uint8_t state);
void cb_time            (uint8_t p, uint32_t milliseconds, uint16_t days);
void cb_sync            (uint8_t p, uint8_t counter);
void cb_init_device     (uint8_t p);
void cb_can_rx          (uint8_t p, uint32_t id, uint8_t *buf, uint8_t buf_len);
void cb_can_tx          (uint8_t p, uint32_t id, uint8_t *buf, uint8_t buf_len);

/*----------------------------------------------------------------------*/
#ifndef _STATIC
    #include <dlfcn.h>
    #if defined(__x86_64__)
        #define LIB_CANOPEN "./lib_canopen_64.so"
    #else
        #define LIB_CANOPEN "./lib_canopen_32.so"
    #endif

INNO_CANOPEN_ENABLE           INNOCANOpenEnable         = NULL;
INNO_CANOPEN_DISABLE          INNOCANOpenDisable        = NULL;
INNO_WRITE_OD                 INNOWriteOD               = NULL;
INNO_CANOPEN_TX               INNOCANOpenTx             = NULL;
INNO_CANOPEN_SET_STATE        INNOCANOpenSetState       = NULL;
INNO_CANOPEN_RQST_STATE       INNOCANOpenRqstState      = NULL;
INNO_CANOPEN_CHECK_NODE_STATE INNOCANOpenCheckNodeState = NULL;
INNO_CANOPEN_GET_VER          INNOCANOpenGetVer         = NULL;

/*----------------------------------------------------------------------*/
void *lib_hd = NULL;

/*----------------------------------------------------------------------*/
int load_so(void)
{
    char *error = NULL;

    lib_hd = dlopen(LIB_CANOPEN, RTLD_LAZY);

    if(!lib_hd)
    {
        printf("%s\n", dlerror());
        return 1;
    }

    INNOCANOpenEnable          = (INNO_CANOPEN_ENABLE)            dlsym(lib_hd, "INNOCANOpenEnable");
    INNOCANOpenDisable         = (INNO_CANOPEN_DISABLE)           dlsym(lib_hd, "INNOCANOpenDisable");
    INNOWriteOD                = (INNO_WRITE_OD)                  dlsym(lib_hd, "INNOWriteOD");
    INNOCANOpenTx              = (INNO_CANOPEN_TX)                dlsym(lib_hd, "INNOCANOpenTx");
    INNOCANOpenSetState        = (INNO_CANOPEN_SET_STATE)         dlsym(lib_hd, "INNOCANOpenSetState");
    INNOCANOpenRqstState       = (INNO_CANOPEN_RQST_STATE)        dlsym(lib_hd, "INNOCANOpenRqstState");
    INNOCANOpenCheckNodeState  = (INNO_CANOPEN_CHECK_NODE_STATE)  dlsym(lib_hd, "INNOCANOpenCheckNodeState");
    INNOCANOpenGetVer          = (INNO_CANOPEN_GET_VER)           dlsym(lib_hd, "INNOCANOpenGetVer");

    if((error = dlerror()) != NULL)
    {
        printf("%s\n", error);
        return 2;
    }

    return 0;
}

#endif /* END: _STATIC */

/*----------------------------------------------------------------------*/
static int handler_ini (void *user, const char *section, const char *name, const char *value)
{
    #define MATCH(s, n)     strcmp(section, s) == 0 && strcmp(name, n) == 0

    char *socket_name = (char *)user;

    /* [setting] section */
    if(MATCH("Device Info", "socket_name"))
    {
        strncpy(socket_name, value, IFNAMSIZ);
        socket_name[IFNAMSIZ] = '\0';
    }
    /* unknown section/name, error */
    else
    {
        return 0;
    }

    return 1;

} /* END: handler_ini() */

/*----------------------------------------------------------------------*/
static void get_ini_setting (char *setup_ini, char *socket_name)
{
    char dest[FILE_PATH_MAX];

    memset(dest, 0, sizeof(dest));

    #if defined(__linux__)
        char path[FILE_PATH_MAX];
        int  len, i;

        memset(path, 0, sizeof(path));

        pid_t pid = getpid();

        sprintf(path, "/proc/%d/exe", pid);
        if(readlink(path, dest, FILE_PATH_MAX) >= 0)
        {
            len = strlen(dest);

            for(i=len; i>0; i--)
            {
                if(dest[i] == '/')
                    break;
                else
                    dest[i] = 0;
            }

            strcat(dest, setup_ini);
        }
    #elif defined(__QNX__)
        char path[FILE_PATH_MAX];
        int  fd, len, i;

        memset(path, 0, sizeof(path));

        pid_t pid = getpid();

        sprintf(path, "/proc/%d/exefile", pid);

        fd = open(path, O_RDONLY);
        if(fd < 0)
        {
            perror("get execute path");
            exit(1);
        }

        len = read(fd, dest, sizeof(dest)-1);
        if(len < 0)
            exit(1);
        
        close(fd);
    
        for(i=len; i>0; i--)
        {
            if(dest[i] == '/')
                break;
            else
                dest[i] = 0;
        }

        strcat(dest, setup_ini);

    #else
        strcat(dest, setup_ini);
    #endif

    if(ini_parse(dest, handler_ini, socket_name) < 0)
    {
        printf("Open ini file failed\n");

        #if defined(_WIN32)
            system("pause");
        #endif

        exit(1);
    }

} /* END: get_ini_setting() */

/*----------------------------------------------------------------------*/
int main (int argc, char *argv[])
{
    int             rtn                       = -1;
    char            *setup_ini                = "setup_master.ini";
    uint8_t         master_port               = 0;
    uint8_t         slave_node_id             = 0x06;
    char            socket_name[IFNAMSIZ + 1] = {0};
    char            version[VER_LEN]          = {0};
    CANOPEN_CB_INFO cb_info                   = {0};
    NSD_INFO        nsd_info                  = {0};
    CAN_TX_INFO     can_tx_info               = {0};

#ifndef _STATIC
    if(load_so())
    {
        printf("Load %s failed!\n", LIB_CANOPEN);
        return 1;
    }
#endif

    /* get socket_name from ini file */
    get_ini_setting(setup_ini, socket_name);
    if(strlen(socket_name) == 0)
    {
        printf("get socket_name from ini file failed!\n");
        goto EXIT;
    }

    INNOCANOpenGetVer(version);
    printf("INNOCANopen ver.%s\n\n", version);

    cb_info.canopen_handle_sdo_abort_cb = cb_handle_sdo_abort;
    cb_info.canopen_handle_sdo_cb       = cb_handle_sdo;
    cb_info.canopen_cob_cb              = cb_cob_handle;
    cb_info.canopen_update_cb           = cb_update_handle;
    cb_info.canopen_set_state_cb        = cb_set_state;
    cb_info.canopen_time_cb             = cb_time;
    cb_info.canopen_sync_cb             = cb_sync;
    cb_info.canopen_init_device_cb      = cb_init_device;
    cb_info.canopen_can_rx_cb           = cb_can_rx;
    cb_info.canopen_can_tx_cb           = cb_can_tx;

    rtn = INNOCANOpenEnable(setup_ini, cb_info);
    if(rtn)
    {
        printf("INNOCANOpenEnable() failed! (rtn=%d)\n", rtn);
        goto EXIT;
    }
    else
    {
        printf("INNOCANOpenEnable() successfully!\n");
    }

    sleep(1);

    printf("\nINNOCANOpenSetState(master_port=%d, slave_node_id=%02X, CMD(ex:op mode))\n", master_port, slave_node_id);
    INNOCANOpenSetState(master_port, slave_node_id, 0x01);   /* CANOPEN_NMT_CMD_START 0x01
                                                                CANOPEN_NMT_CMD_STOP 0x02
                                                                CANOPEN_NMT_CMD_ENTER_PREOP 0x80
                                                                CANOPEN_NMT_CMD_RESET_NODE 0x81
                                                                CANOPEN_NMT_CMD_RESET_COMM 0x82 */

    sleep(1);
    
    printf("\nINNOCANOpenRqstState(master_port=%d, slave_node_id=%02X)(Node guarding)\n", master_port, slave_node_id);
    INNOCANOpenRqstState(master_port, slave_node_id);

    usleep(50*1000);

    printf("INNOCANOpenCheckNodeState(master_port=%d, slave_node_id=%02X), nsd_info)(get node state)\n", master_port, slave_node_id);
    INNOCANOpenCheckNodeState(master_port, slave_node_id, &nsd_info);
    printf("nsd_info=> nodeid:%02X\tstate:%02X\tlastseen_timestamp:%d\n", nsd_info.nodeid, nsd_info.state, nsd_info.lastseen_timestamp);

    sleep(1);

    for(;;)
    {
        uint16_t ix = 0x3000;
        uint8_t sx = 0x00;

        /* for INNOCANOpenTx */
        can_tx_info.id = 0x600 + slave_node_id;
        can_tx_info.buf[0] = (2<<5);
        can_tx_info.buf[1] = ix & 0xFF;
        can_tx_info.buf[2] = ix >> 8;
        can_tx_info.buf[3] = sx;
        can_tx_info.buf[4] = 0;
        can_tx_info.buf[5] = 0;
        can_tx_info.buf[6] = 0;
        can_tx_info.buf[7] = 0;
        can_tx_info.buf_len = 8;
        printf("master INNOCANOpenTx(master_port=%d, can_tx_info)\n", master_port);
        INNOCANOpenTx(master_port, can_tx_info);

        sleep(1);

        if(kbhit() != 0)
            break;
    }

EXIT:

    printf("INNOCANOpenDisable()\n\n");
    INNOCANOpenDisable();
    return 0;

}


/* callback function */
/*----------------------------------------------------------------------*/
void cb_handle_sdo_abort(uint8_t p, uint8_t n, uint16_t index, uint8_t subindex, uint8_t cs, uint32_t abort_code)
{

}

/*----------------------------------------------------------------------*/
void cb_handle_sdo(uint8_t p, uint8_t n, uint16_t index, uint8_t subindex, uint8_t *buf, uint8_t buf_len)
{

}

/*----------------------------------------------------------------------*/
void cb_cob_handle(uint8_t port, uint8_t fcode, uint8_t src, uint8_t *buf, uint8_t buf_len)
{

}

/*----------------------------------------------------------------------*/
void cb_update_handle(int port)
{

}

/*----------------------------------------------------------------------*/
void cb_set_state(uint8_t p, uint8_t state)
{

}

/*----------------------------------------------------------------------*/
void cb_time(uint8_t p, uint32_t milliseconds, uint16_t days)
{

}

/*----------------------------------------------------------------------*/
void cb_sync(uint8_t p, uint8_t counter)
{

}

/*----------------------------------------------------------------------*/
void cb_init_device(uint8_t p)
{

}

/*----------------------------------------------------------------------*/
void cb_can_rx(uint8_t p, uint32_t id, uint8_t *buf, uint8_t buf_len)
{
    if(p == 0)
    {
        printf("path:RX\tid:%03X\tdlc:%d\tdata:", id, buf_len);
        for(uint8_t i=0; i<buf_len; i++)
            printf("[%02X] ", buf[i]);
        printf("\n");
        if(id == 0x586)
        {
            int hour = buf[6];
            int min = buf[5];
            int sec = buf[4];

            printf("current time: %02d:%02d:%02d (from cb_can_rx())\n\n", hour, min, sec);
        }
    }
}

/*----------------------------------------------------------------------*/
void cb_can_tx(uint8_t p, uint32_t id, uint8_t *buf, uint8_t buf_len)
{
    if(p == 0)
    {
        printf("path:TX\tid:%03X\tdlc:%d\tdata:", id, buf_len);
        for(uint8_t i=0; i<buf_len; i++)
            printf("[%02X] ", buf[i]);
        printf("\n");
    }
}

/*----------------------------------------------------------------------*/
/* END: callback function */

